// SPDX-License-Identifier: GPL-2.0

int a(void);
int b(void);
int c(void);
int d(void);
int e(void);
int f(void);
int inc(void);

int main(void)
{
	a();
	b();
	c();
	d();
	e();
	f();
	inc();

	return 0;
}
